@@ -26,15 +26,29 @@ struct _ShumateVectorSymbolLayer
 {
   ShumateVectorLayer parent_instance;
 
+  ShumateVectorExpression *icon_allow_overlap;
   ShumateVectorExpression *icon_anchor;
   ShumateVectorExpression *icon_color;
+  ShumateVectorExpression *icon_ignore_placement;
   ShumateVectorExpression *icon_image;
+  ShumateVectorExpression *icon_opacity;
+  ShumateVectorExpression *icon_optional;
+  ShumateVectorExpression *icon_overlap;
+  ShumateVectorExpression *icon_padding;
+  ShumateVectorExpression *icon_rotate;
   ShumateVectorExpression *icon_rotation_alignment;
   ShumateVectorExpression *icon_size;
   ShumateVectorExpression *text_anchor;
   ShumateVectorExpression *text_field;
+  ShumateVectorExpression *text_letter_spacing;
+  ShumateVectorExpression *text_allow_overlap;
+  ShumateVectorExpression *text_ignore_placement;
   ShumateVectorExpression *text_color;
+  ShumateVectorExpression *text_opacity;
+  ShumateVectorExpression *text_optional;
+  ShumateVectorExpression *text_overlap;
   ShumateVectorExpression *text_size;
+  ShumateVectorExpression *text_transform;
   ShumateVectorExpression *cursor;
   ShumateVectorExpression *text_padding;
   ShumateVectorExpression *text_keep_upright;
@@ -63,9 +77,17 @@ shumate_vector_symbol_layer_create_from_json (JsonObject *object, GError **error
       if (layer->icon_color == NULL)
         return NULL;
 
+      layer->icon_opacity = shumate_vector_expression_from_json (json_object_get_member (paint, "icon-opacity"), error);
+      if (layer->icon_opacity == NULL)
+        return NULL;
+
       layer->text_color = shumate_vector_expression_from_json (json_object_get_member (paint, "text-color"), error);
       if (layer->text_color == NULL)
         return NULL;
+
+      layer->text_opacity = shumate_vector_expression_from_json (json_object_get_member (paint, "text-opacity"), error);
+      if (layer->text_opacity == NULL)
+        return NULL;
     }
 
   if (json_object_has_member (object, "layout"))
@@ -76,10 +98,18 @@ shumate_vector_symbol_layer_create_from_json (JsonObject *object, GError **error
       JsonNode *text_offset_node;
       JsonArray *text_font;
 
+      layer->icon_allow_overlap = shumate_vector_expression_from_json (json_object_get_member (layout, "icon-allow-overlap"), error);
+      if (layer->icon_allow_overlap == NULL)
+        return NULL;
+
       layer->icon_anchor = shumate_vector_expression_from_json (json_object_get_member (layout, "icon-anchor"), error);
       if (layer->icon_anchor == NULL)
         return NULL;
 
+      layer->icon_ignore_placement = shumate_vector_expression_from_json (json_object_get_member (layout, "icon-ignore-placement"), error);
+      if (layer->icon_ignore_placement == NULL)
+        return NULL;
+
       layer->icon_image = shumate_vector_expression_from_json (json_object_get_member (layout, "icon-image"), error);
       if (layer->icon_image == NULL)
         return NULL;
@@ -105,6 +135,22 @@ shumate_vector_symbol_layer_create_from_json (JsonObject *object, GError **error
           layer->icon_offset_y = json_array_get_double_element (icon_offset, 1);
         }
 
+      layer->icon_optional = shumate_vector_expression_from_json (json_object_get_member (layout, "icon-optional"), error);
+      if (layer->icon_optional == NULL)
+        return NULL;
+
+      layer->icon_overlap = shumate_vector_expression_from_json (json_object_get_member (layout, "icon-overlap"), error);
+      if (layer->icon_overlap == NULL)
+        return NULL;
+
+      layer->icon_padding = shumate_vector_expression_from_json (json_object_get_member (layout, "icon-padding"), error);
+      if (layer->icon_padding == NULL)
+        return NULL;
+
+      layer->icon_rotate = shumate_vector_expression_from_json (json_object_get_member (layout, "icon-rotate"), error);
+      if (layer->icon_rotate == NULL)
+        return NULL;
+
       layer->icon_size = shumate_vector_expression_from_json (json_object_get_member (layout, "icon-size"), error);
       if (layer->icon_size == NULL)
         return NULL;
@@ -113,10 +159,26 @@ shumate_vector_symbol_layer_create_from_json (JsonObject *object, GError **error
       if (layer->icon_rotation_alignment == NULL)
         return NULL;
 
+      layer->text_allow_overlap = shumate_vector_expression_from_json (json_object_get_member (layout, "text-allow-overlap"), error);
+      if (layer->text_allow_overlap == NULL)
+        return NULL;
+
       layer->text_field = shumate_vector_expression_from_json (json_object_get_member (layout, "text-field"), error);
       if (layer->text_field == NULL)
         return NULL;
 
+      layer->text_ignore_placement = shumate_vector_expression_from_json (json_object_get_member (layout, "text-ignore-placement"), error);
+      if (layer->text_ignore_placement == NULL)
+        return NULL;
+
+      layer->text_letter_spacing = shumate_vector_expression_from_json (json_object_get_member (layout, "text-letter-spacing"), error);
+      if (layer->text_letter_spacing == NULL)
+        return NULL;
+
+      layer->text_transform = shumate_vector_expression_from_json (json_object_get_member (layout, "text-transform"), error);
+      if (layer->text_transform == NULL)
+        return NULL;
+
       layer->text_anchor = shumate_vector_expression_from_json (json_object_get_member (layout, "text-anchor"), error);
       if (layer->text_anchor == NULL)
         return NULL;
@@ -146,6 +208,14 @@ shumate_vector_symbol_layer_create_from_json (JsonObject *object, GError **error
           layer->text_offset_y = json_array_get_double_element (text_offset, 1);
         }
 
+      layer->text_optional = shumate_vector_expression_from_json (json_object_get_member (layout, "text-optional"), error);
+      if (layer->text_optional == NULL)
+        return NULL;
+
+      layer->text_overlap = shumate_vector_expression_from_json (json_object_get_member (layout, "text-overlap"), error);
+      if (layer->text_overlap == NULL)
+        return NULL;
+
       layer->text_rotation_alignment = shumate_vector_expression_from_json (json_object_get_member (layout, "text-rotation-alignment"), error);
       if (layer->text_rotation_alignment == NULL)
         return NULL;
@@ -211,19 +281,33 @@ shumate_vector_symbol_layer_finalize (GObject *object)
 {
   ShumateVectorSymbolLayer *self = SHUMATE_VECTOR_SYMBOL_LAYER (object);
 
+  g_clear_object (&self->icon_allow_overlap);
   g_clear_object (&self->icon_anchor);
   g_clear_object (&self->icon_color);
+  g_clear_object (&self->icon_ignore_placement);
   g_clear_object (&self->icon_image);
+  g_clear_object (&self->icon_opacity);
+  g_clear_object (&self->icon_optional);
+  g_clear_object (&self->icon_overlap);
+  g_clear_object (&self->icon_padding);
+  g_clear_object (&self->icon_rotate);
   g_clear_object (&self->icon_rotation_alignment);
   g_clear_object (&self->icon_size);
+  g_clear_object (&self->text_allow_overlap);
   g_clear_object (&self->text_field);
+  g_clear_object (&self->text_ignore_placement);
   g_clear_object (&self->text_anchor);
   g_clear_object (&self->text_color);
   g_clear_object (&self->text_size);
   g_clear_object (&self->cursor);
   g_clear_object (&self->text_keep_upright);
+  g_clear_object (&self->text_letter_spacing);
+  g_clear_object (&self->text_opacity);
+  g_clear_object (&self->text_optional);
+  g_clear_object (&self->text_overlap);
   g_clear_object (&self->text_padding);
   g_clear_object (&self->text_rotation_alignment);
+  g_clear_object (&self->text_transform);
   g_clear_object (&self->symbol_sort_key);
   g_clear_object (&self->symbol_placement);
   g_clear_object (&self->symbol_spacing);
@@ -254,80 +338,79 @@ place_point_label (ShumateVectorSymbolDetails *details,
                    double                      y,
                    ShumateVectorRenderScope   *scope)
 {
-  g_ptr_array_add (scope->symbols, create_symbol_info (details, x, y));
+  if (x >= 0 && x < 1 && y >= 0 && y < 1)
+    g_ptr_array_add (scope->symbols, create_symbol_info (details, x, y));
 }
 
 
 static void
 place_line_label (ShumateVectorSymbolDetails *details,
-                  double                      x,
-                  double                      y,
                   ShumateVectorRenderScope   *scope)
 {
   ShumateVectorSymbolInfo *symbol_info;
-  g_autoptr(GPtrArray) lines = shumate_vector_render_scope_get_geometry (scope);
-  guint i, j;
+  GPtrArray *lines;
+  g_autoptr(GPtrArray) simplified_lines = g_ptr_array_new_with_free_func ((GDestroyNotify)shumate_vector_line_string_free);
+  guint i;
+  float spacing = details->symbol_spacing / scope->target_size;
+  float total_length = 0;
+  float remaining_distance;
 
+  if (spacing <= 0)
+    return;
+
+  lines = shumate_vector_render_scope_get_geometry (scope);
+  g_ptr_array_set_free_func (lines, NULL);
   for (i = 0; i < lines->len; i ++)
     {
-      g_autoptr(GPtrArray) split_lines = shumate_vector_line_string_simplify ((ShumateVectorLineString *)lines->pdata[i]);
-
-      for (j = 0; j < split_lines->len; j ++)
+      if (details->text_rotation_alignment == SHUMATE_VECTOR_ALIGNMENT_MAP
+          || details->text_rotation_alignment == SHUMATE_VECTOR_ALIGNMENT_VIEWPORT_GLYPH)
         {
-          g_autoptr(ShumateVectorLineString) linestring = (ShumateVectorLineString *)split_lines->pdata[j];
-          double length = shumate_vector_line_string_length (linestring);
-          float distance = 0;
-          ShumateVectorPointIter iter;
-          float spacing = details->symbol_spacing / scope->target_size;
+          GPtrArray *split_lines = shumate_vector_line_string_simplify ((ShumateVectorLineString *)lines->pdata[i]);
+          g_ptr_array_extend_and_steal (simplified_lines, split_lines);
+        }
+      else
+        g_ptr_array_add (simplified_lines, (ShumateVectorLineString *)lines->pdata[i]);
+    }
+  g_clear_pointer (&lines, g_ptr_array_unref);
 
-          if (spacing <= 0)
-            return;
+  for (i = 0; i < simplified_lines->len; i ++)
+    total_length += shumate_vector_line_string_length ((ShumateVectorLineString *)simplified_lines->pdata[i]);
 
-#if 0
-          /* visualize line simplification */
+  /* Make the spacing even on both sides */
+  remaining_distance = fmod (total_length / 2, spacing);
 
-          if (linestring->n_points > 0)
+  for (i = 0; i < simplified_lines->len; i ++)
+    {
+      ShumateVectorLineString *linestring = (ShumateVectorLineString *)simplified_lines->pdata[i];
+      float distance = 0;
+      ShumateVectorPointIter iter;
+
+      shumate_vector_point_iter_init (&iter, linestring);
+
+      shumate_vector_point_iter_advance (&iter, remaining_distance);
+      distance += remaining_distance;
+
+      while (!shumate_vector_point_iter_is_at_end (&iter))
+        {
+          ShumateVectorPoint point;
+          shumate_vector_point_iter_get_current_point (&iter, &point);
+
+          if (point.x >= 0 && point.x < 1 && point.y >= 0 && point.y < 1)
             {
-              guint k;
-              float scale = scope->scale * scope->target_size;
-
-              cairo_set_source_rgb (scope->cr,
-                                    rand () % 255 / 255.0,
-                                    rand () % 255 / 255.0,
-                                    rand () % 255 / 255.0);
-              cairo_set_line_width (scope->cr, scope->scale);
-
-              cairo_move_to (scope->cr, linestring->points[0].x * scale, linestring->points[0].y * scale);
-              for (k = 1; k < linestring->n_points; k ++)
-                cairo_line_to (scope->cr, linestring->points[k].x * scale, linestring->points[k].y * scale);
-
-              cairo_stroke (scope->cr);
-            }
-#endif
-
-          shumate_vector_point_iter_init (&iter, linestring);
-
-          /* Make the spacing even on both sides */
-          shumate_vector_point_iter_advance (&iter, fmod (length, spacing) / 2);
-          distance += fmod (length, spacing) / 2;
-
-          while (!shumate_vector_point_iter_is_at_end (&iter))
-            {
-              ShumateVectorPoint point;
-              shumate_vector_point_iter_get_current_point (&iter, &point);
-
-              symbol_info = create_symbol_info (details, x, y);
+              symbol_info = create_symbol_info (details, point.x, point.y);
 
               shumate_vector_symbol_info_set_line_points (symbol_info,
                                                           shumate_vector_line_string_copy (linestring),
                                                           distance);
 
               g_ptr_array_add (scope->symbols, symbol_info);
-
-              shumate_vector_point_iter_advance (&iter, spacing);
-              distance += spacing;
             }
+
+          shumate_vector_point_iter_advance (&iter, spacing);
+          distance += spacing;
         }
+
+      remaining_distance = distance - shumate_vector_line_string_length (linestring);
     }
 }
 
@@ -337,15 +420,20 @@ shumate_vector_symbol_layer_render (ShumateVectorLayer *layer, ShumateVectorRend
 {
   ShumateVectorSymbolLayer *self = SHUMATE_VECTOR_SYMBOL_LAYER (layer);
   g_auto(ShumateVectorValue) value = SHUMATE_VECTOR_VALUE_INIT;
+  g_auto(ShumateVectorValue) icon_padding_value = SHUMATE_VECTOR_VALUE_INIT;
+  GPtrArray *icon_padding_array;
   g_autoptr(GPtrArray) text_field = NULL;
   g_autofree char *cursor = NULL;
   g_autofree char *feature_id = NULL;
   ShumateVectorPlacement symbol_placement;
   ShumateVectorAlignment icon_rotation_alignment, text_rotation_alignment;
+  g_autofree char *text_transform = NULL;
   g_autoptr(GHashTable) tags = NULL;
   g_autoptr(ShumateVectorSymbolDetails) details = NULL;
   g_autoptr(ShumateVectorSprite) icon_image = shumate_vector_expression_eval_image (self->icon_image, scope);
   ShumateVectorGeometryType geometry_type = shumate_vector_render_scope_get_geometry_type (scope);
+  VectorTile__Tile__Layer *layer_struct = shumate_vector_reader_iter_get_layer_struct (scope->reader);
+  VectorTile__Tile__Feature *feature = shumate_vector_reader_iter_get_feature_struct (scope->reader);
   double x, y;
 
   shumate_vector_expression_eval (self->text_field, scope, &value);
@@ -363,8 +451,9 @@ shumate_vector_symbol_layer_render (ShumateVectorLayer *layer, ShumateVectorRend
 
       if (strlen (text) > 0)
         {
+          ShumateVectorFormatPart *part;
           text_field = g_ptr_array_new_with_free_func ((GDestroyNotify)shumate_vector_format_part_free);
-          ShumateVectorFormatPart *part = g_new0 (ShumateVectorFormatPart, 1);
+          part = g_new0 (ShumateVectorFormatPart, 1);
           part->string = g_steal_pointer (&text);
           g_ptr_array_add (text_field, part);
         }
@@ -373,6 +462,22 @@ shumate_vector_symbol_layer_render (ShumateVectorLayer *layer, ShumateVectorRend
   if (text_field == NULL && icon_image == NULL)
     return;
 
+  text_transform = shumate_vector_expression_eval_string (self->text_transform, scope, NULL);
+  if (text_field != NULL && text_transform != NULL)
+    {
+      for (int i = 0; i < text_field->len; i ++)
+        {
+          ShumateVectorFormatPart *part = g_ptr_array_index (text_field, i);
+          if (part->string != NULL)
+            {
+              if (g_strcmp0 (text_transform, "uppercase") == 0)
+                part->string = g_utf8_strup (part->string, -1);
+              else if (g_strcmp0 (text_transform, "lowercase") == 0)
+                part->string = g_utf8_strdown (part->string, -1);
+            }
+        }
+    }
+
   symbol_placement = shumate_vector_expression_eval_placement (self->symbol_placement, scope);
   if (geometry_type == SHUMATE_VECTOR_GEOMETRY_POINT && symbol_placement != SHUMATE_VECTOR_PLACEMENT_POINT)
     return;
@@ -393,7 +498,7 @@ shumate_vector_symbol_layer_render (ShumateVectorLayer *layer, ShumateVectorRend
         : SHUMATE_VECTOR_ALIGNMENT_MAP;
     }
 
-  feature_id = g_strdup_printf ("%ld", scope->feature->id);
+  feature_id = g_strdup_printf ("%ld", feature->id);
   cursor = shumate_vector_expression_eval_string (self->cursor, scope, NULL);
 
   tags = shumate_vector_render_scope_create_tag_table (scope);
@@ -402,12 +507,17 @@ shumate_vector_symbol_layer_render (ShumateVectorLayer *layer, ShumateVectorRend
   *details = (ShumateVectorSymbolDetails) {
     .ref_count = 1,
     .layer = g_strdup (shumate_vector_layer_get_id (layer)),
-    .source_layer = g_strdup (scope->layer->name),
+    .source_layer = g_strdup (layer_struct->name),
     .feature_id = g_strdup (feature_id),
     .tags = g_hash_table_ref (tags),
 
     .icon_anchor = shumate_vector_expression_eval_anchor (self->icon_anchor, scope),
+    .icon_ignore_placement = shumate_vector_expression_eval_boolean (self->icon_ignore_placement, scope, FALSE),
     .icon_image = g_steal_pointer (&icon_image),
+    .icon_opacity = CLAMP (shumate_vector_expression_eval_number (self->icon_opacity, scope, 1.0), 0.0, 1.0),
+    .icon_optional = shumate_vector_expression_eval_boolean (self->icon_optional, scope, FALSE),
+    .icon_overlap = shumate_vector_expression_eval_overlap (self->icon_overlap, self->icon_allow_overlap, scope),
+    .icon_rotate = shumate_vector_expression_eval_number (self->icon_rotate, scope, 0.0) * G_PI / 180,
     .icon_rotation_alignment = icon_rotation_alignment,
     .icon_size = shumate_vector_expression_eval_number (self->icon_size, scope, 1.0),
     .icon_offset_x = self->icon_offset_x,
@@ -416,6 +526,11 @@ shumate_vector_symbol_layer_render (ShumateVectorLayer *layer, ShumateVectorRend
     .formatted_text = g_steal_pointer (&text_field),
 
     .text_anchor = shumate_vector_expression_eval_anchor (self->text_anchor, scope),
+    .text_ignore_placement = shumate_vector_expression_eval_boolean (self->text_ignore_placement, scope, FALSE),
+    .text_letter_spacing = shumate_vector_expression_eval_number (self->text_letter_spacing, scope, 0.0),
+    .text_opacity = shumate_vector_expression_eval_number (self->text_opacity, scope, 1.0),
+    .text_optional = shumate_vector_expression_eval_boolean (self->text_optional, scope, FALSE),
+    .text_overlap = shumate_vector_expression_eval_overlap (self->text_overlap, self->text_allow_overlap, scope),
     .text_size = shumate_vector_expression_eval_number (self->text_size, scope, 16.0),
     .text_padding = shumate_vector_expression_eval_number (self->text_padding, scope, 2.0),
     .text_font = g_strdup (self->text_fonts),
@@ -441,12 +556,36 @@ shumate_vector_symbol_layer_render (ShumateVectorLayer *layer, ShumateVectorRend
   details->text_color = SHUMATE_VECTOR_COLOR_BLACK;
   shumate_vector_expression_eval_color (self->text_color, scope, &details->text_color);
 
+  shumate_vector_expression_eval (self->icon_padding, scope, &icon_padding_value);
+  if ((icon_padding_array = shumate_vector_value_get_array (&icon_padding_value)) != NULL)
+    {
+      if (icon_padding_array->len == 0 || !shumate_vector_value_get_number (g_ptr_array_index (icon_padding_array, 0), &details->icon_padding_top))
+        details->icon_padding_top = 2;
+
+      if (icon_padding_array->len <= 1 || !shumate_vector_value_get_number (g_ptr_array_index (icon_padding_array, 1), &details->icon_padding_right))
+        details->icon_padding_right = details->icon_padding_top;
+
+      if (icon_padding_array->len <= 2 || !shumate_vector_value_get_number (g_ptr_array_index (icon_padding_array, 2), &details->icon_padding_bottom))
+        details->icon_padding_bottom = details->icon_padding_top;
+
+      if (icon_padding_array->len <= 3 || !shumate_vector_value_get_number (g_ptr_array_index (icon_padding_array, 3), &details->icon_padding_left))
+        details->icon_padding_left = details->icon_padding_right;
+    }
+  else
+    {
+      if (!shumate_vector_value_get_number (&icon_padding_value, &details->icon_padding_top))
+        details->icon_padding_top = 2;
+
+      details->icon_padding_right = details->icon_padding_top;
+      details->icon_padding_bottom = details->icon_padding_top;
+      details->icon_padding_left = details->icon_padding_top;
+    }
+
   switch (shumate_vector_render_scope_get_geometry_type (scope))
     {
     case SHUMATE_VECTOR_GEOMETRY_POINT:
         {
           g_autoptr(GPtrArray) geometry = shumate_vector_render_scope_get_geometry (scope);
-          g_ptr_array_set_free_func (geometry, (GDestroyNotify)shumate_vector_line_string_free);
 
           for (int i = 0; i < geometry->len; i ++)
             {
@@ -456,8 +595,6 @@ shumate_vector_symbol_layer_render (ShumateVectorLayer *layer, ShumateVectorRend
                 {
                   x = string->points[j].x;
                   y = string->points[j].y;
-                  if (x < 0 || x >= 1 || y < 0 || y >= 1)
-                    break;
                   place_point_label (details, x, y, scope);
                 }
             }
@@ -468,34 +605,37 @@ shumate_vector_symbol_layer_render (ShumateVectorLayer *layer, ShumateVectorRend
     case SHUMATE_VECTOR_GEOMETRY_LINESTRING:
     case SHUMATE_VECTOR_GEOMETRY_POLYGON:
       shumate_vector_render_scope_get_geometry_center (scope, &x, &y);
-      if (x < 0 || x >= 1 || y < 0 || y >= 1)
-        /* Tiles usually include a bit of margin. Don't include symbols that are
-         * covered by a different tile. */
-        break;
 
       if (symbol_placement == SHUMATE_VECTOR_PLACEMENT_LINE
           || symbol_placement == SHUMATE_VECTOR_PLACEMENT_LINE_CENTER)
-        place_line_label (details, x, y, scope);
+        place_line_label (details, scope);
       else
         {
           if (shumate_vector_render_scope_get_geometry_type (scope) == SHUMATE_VECTOR_GEOMETRY_LINESTRING)
             {
-              // Place at the middle point of the linestring
+              double distance = 0;
               g_autoptr(GPtrArray) geometry = shumate_vector_render_scope_get_geometry (scope);
-              g_ptr_array_set_free_func (geometry, (GDestroyNotify)shumate_vector_line_string_free);
+
+              for (int i = 0; i < geometry->len; i ++)
+                distance += shumate_vector_line_string_length (g_ptr_array_index (geometry, i)) / 2;
 
               for (int i = 0; i < geometry->len; i ++)
                 {
                   ShumateVectorLineString *linestring = g_ptr_array_index (geometry, i);
-                  double length = shumate_vector_line_string_length (linestring);
                   ShumateVectorPointIter iter;
-                  ShumateVectorPoint point;
 
                   shumate_vector_point_iter_init (&iter, linestring);
-                  shumate_vector_point_iter_advance (&iter, length / 2);
-                  shumate_vector_point_iter_get_current_point (&iter, &point);
+                  shumate_vector_point_iter_advance (&iter, distance);
 
-                  place_point_label (details, point.x, point.y, scope);
+                  if (shumate_vector_point_iter_is_at_end (&iter))
+                    distance -= shumate_vector_line_string_length (linestring);
+                  else
+                    {
+                      ShumateVectorPoint point;
+                      shumate_vector_point_iter_get_current_point (&iter, &point);
+                      place_point_label (details, point.x, point.y, scope);
+                      break;
+                    }
                 }
             }
           else
