@@ -35,7 +35,7 @@ struct _ShumateVectorSymbol
   GArray *glyphs;
 
   GskRenderNode *glyphs_node;
-  int layout_width, layout_height;
+  int layout_width, layout_height, baseline, layout_y;
 
   graphene_rect_t bounds;
   double x, y;
@@ -43,6 +43,9 @@ struct _ShumateVectorSymbol
   ShumateVectorPoint midpoint;
   double midpoint_angle;
   double line_length;
+
+  uint8_t show_text : 1;
+  uint8_t show_icon : 1;
 };
 
 G_DEFINE_TYPE (ShumateVectorSymbol, shumate_vector_symbol, GTK_TYPE_WIDGET)
@@ -130,6 +133,7 @@ shumate_vector_symbol_constructed (GObject *object)
       g_autoptr(GString) string = g_string_new ("");
       PangoAttribute *attr;
       gboolean any_images = FALSE;
+      PangoRectangle ink_rect;
 
       if (self->symbol_info->details->text_font != NULL)
         {
@@ -138,6 +142,9 @@ shumate_vector_symbol_constructed (GObject *object)
           pango_attr_list_insert (attrs, attr);
         }
 
+      attr = pango_attr_letter_spacing_new (self->symbol_info->details->text_letter_spacing * self->symbol_info->details->text_size * PANGO_SCALE);
+      pango_attr_list_insert (attrs, attr);
+
       attr = pango_attr_foreground_new (self->symbol_info->details->text_color.red * 65535,
                                         self->symbol_info->details->text_color.green * 65535,
                                         self->symbol_info->details->text_color.blue * 65535);
@@ -159,18 +166,36 @@ shumate_vector_symbol_constructed (GObject *object)
             {
               int width = shumate_vector_sprite_get_width (part->sprite);
               int height = shumate_vector_sprite_get_height (part->sprite);
-              PangoRectangle rect = {
-                .x = 0,
+              /* For shapes, since we're overriding the ink and logical rects
+                 of the glyph, we have to take letter spacing into account ourselves */
+              double spacing = self->symbol_info->details->text_letter_spacing * self->symbol_info->details->text_size;
+
+              PangoRectangle ink_rect = {
+                .x = spacing / 2 * PANGO_SCALE,
                 .y = -height * PANGO_SCALE,
                 .width = width * PANGO_SCALE,
                 .height = height * PANGO_SCALE,
               };
+              PangoRectangle logical_rect = {
+                .x = 0,
+                .y = -height * PANGO_SCALE,
+                .width = (width + spacing) * PANGO_SCALE,
+                .height = height * PANGO_SCALE,
+              };
 
-              attr = pango_attr_shape_new_with_data (&rect, &rect, g_object_ref (part->sprite), (PangoAttrDataCopyFunc)g_object_ref, g_object_unref);
+              attr = pango_attr_shape_new_with_data (&ink_rect, &logical_rect, g_object_ref (part->sprite), (PangoAttrDataCopyFunc)g_object_ref, g_object_unref);
               attr->start_index = string->len;
               attr->end_index = string->len + strlen ("\uFFFC");
               pango_attr_list_insert (attrs, attr);
 
+              if (spacing != 0)
+                {
+                  attr = pango_attr_letter_spacing_new (0);
+                  attr->start_index = string->len;
+                  attr->end_index = string->len + strlen ("\uFFFC");
+                  pango_attr_list_insert (attrs, attr);
+                }
+
               g_string_append (string, "\uFFFC");
 
               any_images = TRUE;
@@ -207,9 +232,11 @@ shumate_vector_symbol_constructed (GObject *object)
       text = g_string_free (g_steal_pointer (&string), FALSE);
       pango_layout_set_text (layout, text, -1);
 
-      pango_layout_get_size (layout, &self->layout_width, &self->layout_height);
-      self->layout_width /= PANGO_SCALE;
-      self->layout_height /= PANGO_SCALE;
+      pango_layout_get_pixel_extents (layout, &ink_rect, NULL);
+      self->layout_width = ink_rect.width;
+      self->layout_height = ink_rect.height;
+      self->layout_y = ink_rect.y;
+      self->baseline = pango_layout_get_baseline (layout) / (double) PANGO_SCALE;
 
       if ((self->symbol_info->details->text_rotation_alignment == SHUMATE_VECTOR_ALIGNMENT_MAP
           || self->symbol_info->details->text_rotation_alignment == SHUMATE_VECTOR_ALIGNMENT_VIEWPORT_GLYPH)
@@ -363,17 +390,17 @@ shumate_vector_symbol_contains (GtkWidget *widget,
                                 double     y)
 {
   GtkWidget *parent;
-  GtkAllocation alloc;
+  graphene_rect_t alloc;
 
   if (x < 0 || y < 0)
     return FALSE;
 
-  gtk_widget_get_allocation (widget, &alloc);
-
-  if (x > alloc.width || y > alloc.height)
+  parent = gtk_widget_get_parent (widget);
+  if (!gtk_widget_compute_bounds (widget, parent, &alloc))
     return FALSE;
 
-  parent = gtk_widget_get_parent (widget);
+  if (x > alloc.size.width || y > alloc.size.height)
+    return FALSE;
 
   if (SHUMATE_IS_VECTOR_SYMBOL_CONTAINER (parent))
     {
@@ -381,8 +408,8 @@ shumate_vector_symbol_contains (GtkWidget *widget,
         shumate_vector_symbol_container_get_collision ((ShumateVectorSymbolContainer *) (parent));
 
       return shumate_vector_collision_query_point (collision,
-                                                   alloc.x + x + collision->delta_x,
-                                                   alloc.y + y + collision->delta_y,
+                                                   alloc.origin.x + x + collision->delta_x,
+                                                   alloc.origin.y + y + collision->delta_y,
                                                    widget);
     }
 
@@ -543,7 +570,7 @@ shumate_vector_symbol_snapshot (GtkWidget   *widget,
       ShumateViewport *viewport = shumate_layer_get_viewport (SHUMATE_LAYER (parent));
       ShumateMapSource *map_source = shumate_vector_symbol_container_get_map_source (SHUMATE_VECTOR_SYMBOL_CONTAINER (parent));
       double zoom_level = shumate_viewport_get_zoom_level (viewport);
-      tile_size_for_zoom = shumate_map_source_get_tile_size (map_source) * powf (2, zoom_level - self->symbol_info->details->tile_zoom_level);
+      tile_size_for_zoom = shumate_map_source_get_tile_size (map_source) * pow (2, zoom_level - self->symbol_info->details->tile_zoom_level);
       rotation = shumate_viewport_get_rotation (viewport);
     }
 
@@ -553,7 +580,7 @@ shumate_vector_symbol_snapshot (GtkWidget   *widget,
   gtk_snapshot_translate (snapshot, &GRAPHENE_POINT_INIT (self->x - self->bounds.origin.x,
                                                           self->y - self->bounds.origin.y));
 
-  if (self->symbol_info->details->icon_image)
+  if (self->show_icon && self->symbol_info->details->icon_image && self->symbol_info->details->icon_opacity > 0.0)
     {
       double angle = 0;
       double icon_width = shumate_vector_sprite_get_width (self->symbol_info->details->icon_image)
@@ -573,6 +600,8 @@ shumate_vector_symbol_snapshot (GtkWidget   *widget,
       else
         angle = -rotation;
 
+      angle += self->symbol_info->details->icon_rotate;
+
       gtk_snapshot_save (snapshot);
 
       gtk_snapshot_rotate (snapshot, rotation * 180 / G_PI);
@@ -587,6 +616,10 @@ shumate_vector_symbol_snapshot (GtkWidget   *widget,
                                 -icon_width / 2.0 + offset_x,
                                 -icon_height / 2.0 + offset_y
                               ));
+
+      if (self->symbol_info->details->icon_opacity < 1.0)
+        gtk_snapshot_push_opacity (snapshot, self->symbol_info->details->icon_opacity);
+
       gtk_symbolic_paintable_snapshot_symbolic (
         GTK_SYMBOLIC_PAINTABLE (self->symbol_info->details->icon_image),
         snapshot,
@@ -595,10 +628,14 @@ shumate_vector_symbol_snapshot (GtkWidget   *widget,
         &self->symbol_info->details->icon_color,
         1
       );
+
+      if (self->symbol_info->details->icon_opacity < 1.0)
+        gtk_snapshot_pop (snapshot);
+
       gtk_snapshot_restore (snapshot);
     }
 
-  if (self->glyphs)
+  if (self->show_text && self->glyphs && self->symbol_info->details->text_opacity > 0.0)
     {
       double length = self->layout_width / tile_size_for_zoom;
       double start_pos = MAX (0, self->symbol_info->line_position - (length / 2.0));
@@ -624,6 +661,9 @@ shumate_vector_symbol_snapshot (GtkWidget   *widget,
       gtk_snapshot_save (snapshot);
       gtk_snapshot_rotate (snapshot, rotation * 180 / G_PI);
 
+      if (self->symbol_info->details->text_opacity < 1.0)
+        gtk_snapshot_push_opacity (snapshot, self->symbol_info->details->text_opacity);
+
       for (int i = 0; i < self->glyphs->len; i ++)
         {
           Glyph *glyph = &((Glyph *)self->glyphs->data)[i];
@@ -653,7 +693,7 @@ shumate_vector_symbol_snapshot (GtkWidget   *widget,
               gtk_snapshot_rotate (snapshot, angle * 180 / G_PI);
               gtk_snapshot_translate (snapshot,
                                       &GRAPHENE_POINT_INIT (-glyph->width / 2.0,
-                                                            self->symbol_info->details->text_size / 2.0));
+                                                            self->baseline - self->layout_y - self->layout_height / 2.0));
 
               if (glyph->node != NULL)
                 gtk_snapshot_append_node (snapshot, glyph->node);
@@ -678,9 +718,12 @@ shumate_vector_symbol_snapshot (GtkWidget   *widget,
             shumate_vector_point_iter_advance (&iter, glyph->width / tile_size_for_zoom);
         }
 
+      if (self->symbol_info->details->text_opacity < 1.0)
+        gtk_snapshot_pop (snapshot);
+
       gtk_snapshot_restore (snapshot);
     }
-  else if (self->glyphs_node)
+  else if (self->show_text && self->glyphs_node && self->symbol_info->details->text_opacity > 0.0)
     {
       double angle = 0;
       double offset_x = self->symbol_info->details->text_offset_x * self->symbol_info->details->text_size;
@@ -695,6 +738,9 @@ shumate_vector_symbol_snapshot (GtkWidget   *widget,
 
       gtk_snapshot_save (snapshot);
 
+      if (self->symbol_info->details->text_opacity < 1)
+        gtk_snapshot_push_opacity (snapshot, self->symbol_info->details->text_opacity);
+
       gtk_snapshot_rotate (snapshot, rotation * 180 / G_PI);
       gtk_snapshot_translate (snapshot,
                               &GRAPHENE_POINT_INIT (
@@ -703,9 +749,15 @@ shumate_vector_symbol_snapshot (GtkWidget   *widget,
                               ));
       gtk_snapshot_rotate (snapshot, angle * 180 / G_PI);
 
-      gtk_snapshot_translate (snapshot, &GRAPHENE_POINT_INIT (-self->layout_width / 2.0 + offset_x,
-                                                              -self->layout_height / 2.0 + offset_y));
+      gtk_snapshot_translate (snapshot, &GRAPHENE_POINT_INIT (
+                                          -self->layout_width / 2.0 + offset_x,
+                                          -self->layout_y - self->layout_height / 2.0 + offset_y
+                                        ));
       gtk_snapshot_append_node (snapshot, self->glyphs_node);
+
+      if (self->symbol_info->details->text_opacity < 1)
+        gtk_snapshot_pop (snapshot);
+
       gtk_snapshot_restore (snapshot);
     }
 
@@ -758,10 +810,13 @@ on_clicked (ShumateVectorSymbol *self,
             double               y,
             GtkGestureClick     *click)
 {
-  g_autoptr(ShumateSymbolEvent) event = shumate_symbol_event_new (self->symbol_info->details->layer,
-                                                                  self->symbol_info->details->source_layer,
-                                                                  self->symbol_info->details->feature_id,
-                                                                  self->symbol_info->details->tags);
+  g_autoptr(ShumateSymbolEvent) event =
+    shumate_symbol_event_new_with_n_press (self->symbol_info->details->layer,
+                                           self->symbol_info->details->source_layer,
+                                           self->symbol_info->details->feature_id,
+                                           self->symbol_info->details->tags,
+                                           n_press);
+
   g_signal_emit (self, signals[CLICKED], 0, event);
 }
 
@@ -814,6 +869,7 @@ shumate_vector_symbol_calculate_collision (ShumateVectorSymbol    *self,
   ShumateVectorPointIter iter;
   ShumateVectorPoint point;
   ShumateVectorPoint midpoint = {0};
+  int save = 0;
 
   shumate_vector_collision_rollback_pending (collision, 0);
 
@@ -835,11 +891,16 @@ shumate_vector_symbol_calculate_collision (ShumateVectorSymbol    *self,
       g_assert (self->symbol_info->line != NULL);
 
       if (length > line_length - start_pos)
-        return FALSE;
+        {
+          self->show_text = FALSE;
+          return FALSE;
+        }
 
       shumate_vector_point_iter_init (&iter, self->symbol_info->line);
       shumate_vector_point_iter_advance (&iter, start_pos);
 
+      self->show_text = TRUE;
+
       do
         {
           double xextent = MIN (length, shumate_vector_point_iter_get_segment_length (&iter) - iter.distance) * tile_size_for_zoom / 2;
@@ -862,11 +923,22 @@ shumate_vector_symbol_calculate_collision (ShumateVectorSymbol    *self,
             xextent + self->symbol_info->details->text_padding,
             yextent + self->symbol_info->details->text_padding,
             rotation + shumate_vector_point_iter_get_current_angle (&iter),
+            self->symbol_info->details->text_overlap,
+            self->symbol_info->details->text_ignore_placement,
             self
           );
 
           if (!check)
-            return FALSE;
+            {
+              if (self->symbol_info->details->text_optional)
+                {
+                  shumate_vector_collision_rollback_pending (collision, save);
+                  self->show_text = FALSE;
+                  break;
+                }
+              else
+                return FALSE;
+            }
         }
       while ((length -= shumate_vector_point_iter_next_segment (&iter)) > 0);
     }
@@ -884,6 +956,8 @@ shumate_vector_symbol_calculate_collision (ShumateVectorSymbol    *self,
 
       rotate_around_center (&offset_x, &offset_y, angle);
 
+      self->show_text = TRUE;
+
       check = shumate_vector_collision_check (
         collision,
         x + midpoint.x + offset_x,
@@ -891,13 +965,25 @@ shumate_vector_symbol_calculate_collision (ShumateVectorSymbol    *self,
         self->layout_width / 2.0 + self->symbol_info->details->text_padding,
         yextent + self->symbol_info->details->text_padding,
         angle,
+        self->symbol_info->details->text_overlap,
+        self->symbol_info->details->text_ignore_placement,
         self
       );
 
       if (!check)
-        return FALSE;
+        {
+          if (self->symbol_info->details->text_optional)
+            {
+              shumate_vector_collision_rollback_pending (collision, save);
+              self->show_text = FALSE;
+            }
+          else
+            return FALSE;
+        }
     }
 
+  save = shumate_vector_collision_save_pending (collision);
+
   if (self->symbol_info->details->icon_image)
     {
       float angle = 0;
@@ -909,27 +995,44 @@ shumate_vector_symbol_calculate_collision (ShumateVectorSymbol    *self,
       double offset_x = self->symbol_info->details->icon_offset_x * self->symbol_info->details->icon_size;
       double offset_y = self->symbol_info->details->icon_offset_y * self->symbol_info->details->icon_size;
 
-      add_anchor_offset (self->symbol_info->details->icon_anchor, &offset_x, &offset_y, icon_width, icon_height);
+      self->show_icon = TRUE;
 
-      rotate_around_center (&offset_x, &offset_y, angle);
+      add_anchor_offset (self->symbol_info->details->icon_anchor, &offset_x, &offset_y, icon_width, icon_height);
 
       if (self->symbol_info->details->icon_rotation_alignment == SHUMATE_VECTOR_ALIGNMENT_MAP)
         angle = rotation + self->midpoint_angle;
+      angle += self->symbol_info->details->icon_rotate;
+      offset_x += (self->symbol_info->details->icon_padding_right - self->symbol_info->details->icon_padding_left) / 2.0;
+      offset_y += (self->symbol_info->details->icon_padding_bottom - self->symbol_info->details->icon_padding_top) / 2.0;
+      rotate_around_center (&offset_x, &offset_y, angle);
 
       check = shumate_vector_collision_check (
         collision,
         x + midpoint.x + offset_x,
         y + midpoint.y + offset_y,
-        icon_width / 2,
-        icon_height / 2,
+        (icon_width + self->symbol_info->details->icon_padding_left + self->symbol_info->details->icon_padding_right) / 2 ,
+        (icon_height + self->symbol_info->details->icon_padding_top + self->symbol_info->details->icon_padding_bottom) / 2,
         angle,
+        self->symbol_info->details->icon_overlap,
+        self->symbol_info->details->icon_ignore_placement,
         self
       );
 
       if (!check)
-        return FALSE;
+        {
+          if (self->symbol_info->details->icon_optional)
+            {
+              shumate_vector_collision_rollback_pending (collision, save);
+              self->show_icon = FALSE;
+            }
+          else
+            return FALSE;
+        }
     }
 
+  if (!self->show_icon && !self->show_text)
+    return FALSE;
+
   shumate_vector_collision_commit_pending (collision, &self->bounds);
   *bounds_out = self->bounds;
   self->x = x;
